@@ -1,70 +1,58 @@
+# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
+# SPDX-License-Identifier: BSD-4-Clause
 from __future__ import annotations
 
 import re
 
+from .tokenizing import LineInfo
 
-class ParseException(Exception):
+
+class TatSuException(Exception):
     pass
 
 
-# alias
-TatSuException = ParseException
+class ParseException(TatSuException):
+    pass
 
 
 class OptionSucceeded(ParseException):
     pass
 
 
-class GrammarError(ParseException):
-    pass
-
-
-class SemanticError(ParseException):
-    pass
-
-
-class CodegenError(ParseException):
-    pass
-
-
-class MissingSemanticFor(SemanticError):
-    pass
-
-
 class ParseError(ParseException):
     pass
 
 
-class FailedSemantics(ParseError):
+class GrammarError(ParseError):
     pass
 
 
-class FailedKeywordSemantics(FailedSemantics):
+class CodegenError(ParseError):
     pass
 
 
-class NoParseInfo(ParseException):
+class FailedSemantics(ParseException):
     pass
 
 
-class FailedParse(ParseError):
-    def __init__(self, tokenizer, stack, item):
-        stack = list(stack)  # NOTE: can't pass through multiprocessing if generator
-        # note: pass all arguments to super() to avoid pickling problems
-        #   https://stackoverflow.com/questions/27993567/
-        super().__init__(tokenizer, stack, item)
+class FailedParse(ParseException):
+    def __init__(self, lineinfo: LineInfo, stack: list[str], msg: str):
+        # NOTE:
+        #   Pass all arguments to super() to avoid pickling problems
+        #       https://stackoverflow.com/questions/27993567/
+        super().__init__(lineinfo, stack, msg)
 
-        self.tokenizer = tokenizer
+        self.lineinfo = lineinfo
         self.stack = stack
-        self.pos = tokenizer.pos
-        self.item = item
+        self.msg = msg
+        self.pos = lineinfo.end
 
     @property
     def message(self):
-        return self.item
+        return self.msg
 
     def __str__(self):
-        info = self.tokenizer.line_info(self.pos)
+        info = self.lineinfo
         template = '{}({}:{}) {} :\n{}\n{}^\n{}'
         text = info.text.rstrip()
         leading = re.sub(r'[^\t]', ' ', text)[: info.col]
@@ -93,13 +81,7 @@ class FailedToken(FailedParse):
 
 
 class FailedPattern(FailedParse):
-    def __init__(self, tokenizer, stack, pattern):
-        super().__init__(tokenizer, stack, pattern)
-        self.pattern = pattern
-
-    @property
-    def message(self):
-        return f'expecting /{self.pattern}/'
+    pass
 
 
 class FailedRef(FailedParse):
@@ -109,13 +91,13 @@ class FailedRef(FailedParse):
 
     @property
     def message(self):
-        return f"could not resolve reference to rule '{self.name}'"
+        return f"Could not resolve reference to rule '{self.name}'"
 
 
 class FailedChoice(FailedParse):
     @property
     def message(self):
-        return 'no viable option'
+        return 'No viable option'
 
 
 class FailedLookahead(FailedParse):
@@ -134,5 +116,8 @@ class FailedExpectingEndOfText(FailedParse):
     pass
 
 
-class FailedKeyword(FailedParse):
+class KeywordError(FailedParse):
     pass
+
+
+FailedKeywordSemantics = KeywordError
